package ru.dbotthepony.kommons.math

/**
 * Calculates n!
 */
fun factorial(n: Int): Int {
	require(n >= 0) { "Tried to compute ($n)!" }

	when (n) {
		0, 1 -> return 1
		2 -> return 2
		3 -> return 6
	}

	var value = 6

	for (j in 4..n) {
		value *= j
	}

	return value
}

/**
 * Calculates n!
 */
fun factorial(n: Long): Long {
	require(n >= 0L) { "Tried to compute ($n)!" }

	when (n) {
		0L, 1L -> return 1L
		2L -> return 2L
		3L -> return 6L
	}

	var value = 6L

	for (j in 4L..n) {
		value *= j
	}

	return value
}

/**
 * Calculates n!
 *
 * For higher order of [n] this is prone to rounding errors.
 */
fun factorialFloat(n: Long): Float {
	require(n >= 0L) { "Tried to compute ($n)!" }

	when (n) {
		0L, 1L -> return 1f
		2L -> return 2f
		3L -> return 6f
	}

	var value = 6f

	for (j in 4L..n) {
		value *= j
	}

	return value
}

/**
 * Calculates n!
 *
 * For higher order of [n] this is prone to rounding errors.
 */
fun factorialDouble(n: Long): Double {
	require(n >= 0L) { "Tried to compute ($n)!" }

	when (n) {
		0L, 1L -> return 1.0
		2L -> return 2.0
		3L -> return 6.0
	}

	var value = 6.0

	for (j in 4L..n) {
		value *= j
	}

	return value
}

/**
 * Calculates n!
 *
 * For higher order of [n] this is prone to rounding errors.
 */
fun factorialFloat(n: Int): Float {
	require(n >= 0L) { "Tried to compute ($n)!" }

	when (n) {
		0, 1 -> return 1f
		2 -> return 2f
		3 -> return 6f
	}

	var value = 6f

	for (j in 4..n) {
		value *= j
	}

	return value
}

/**
 * Calculates n!
 *
 * For higher order of [n] this is prone to rounding errors.
 */
fun factorialDouble(n: Int): Double {
	require(n >= 0L) { "Tried to compute ($n)!" }

	when (n) {
		0, 1 -> return 1.0
		2 -> return 2.0
		3 -> return 6.0
	}

	var value = 6.0

	for (j in 4..n) {
		value *= j
	}

	return value
}

/**
 * Calculates binomial coefficient of n by k
 */
fun binomial(n: Int, k: Int): Int {
	return factorial(n) / (factorial(k) * factorial(n - k))
}

/**
 * Calculates binomial coefficient of n by k
 */
fun binomial(n: Long, k: Long): Long {
	return factorial(n) / (factorial(k) * factorial(n - k))
}

/**
 * Calculates binomial coefficient of n by k.
 *
 * For higher order of [n] or [k] this is prone to rounding errors.
 */
fun binomialFloat(n: Int, k: Int): Float {
	return factorialFloat(n) / (factorialFloat(k) * factorialFloat(n - k))
}

/**
 * Calculates binomial coefficient of n by k.
 *
 * For higher order of [n] or [k] this is prone to rounding errors.
 */
fun binomialFloat(n: Long, k: Long): Float {
	return factorialFloat(n) / (factorialFloat(k) * factorialFloat(n - k))
}

/**
 * Calculates binomial coefficient of n by k.
 *
 * For higher order of [n] or [k] this is prone to rounding errors.
 */
fun binomialDouble(n: Int, k: Int): Double {
	return factorialDouble(n) / (factorialDouble(k) * factorialDouble(n - k))
}

/**
 * Calculates binomial coefficient of n by k
 *
 * For higher order of [n] or [k] this is prone to rounding errors.
 */
fun binomialDouble(n: Long, k: Long): Double {
	return factorialDouble(n) / (factorialDouble(k) * factorialDouble(n - k))
}

/**
 * Calculates a row in pascal triangle, returning result as [IntArray].
 */
fun pascalTriangleRow(row: Int): IntArray {
	val built = IntArray(row + 1)

	for (k in 0 .. row) {
		built[k] = binomial(row, k)
	}

	return built
}

/**
 * Calculates a row in pascal triangle, returning result as [LongArray].
 */
fun pascalTriangleRowLong(row: Int): LongArray {
	val built = LongArray(row + 1)
	val rowL = row.toLong()

	for (k in 0 .. row) {
		built[k] = binomial(rowL, k.toLong())
	}

	return built
}

/**
 * Calculates a row in pascal triangle, returning result as [FloatArray].
 *
 * High orders of polynomials are numerically unstable and prone to increasing rounding errors.
 */
fun pascalTriangleRowFloat(row: Int): FloatArray {
	val built = FloatArray(row + 1)
	val rowL = row.toLong()

	for (k in 0 .. row) {
		built[k] = binomialFloat(rowL, k.toLong())
	}

	return built
}

/**
 * Calculates a row in pascal triangle, returning result as [DoubleArray].
 *
 * High orders of polynomials are numerically unstable and prone to increasing rounding errors.
 */
fun pascalTriangleRowDouble(row: Int): DoubleArray {
	val built = DoubleArray(row + 1)
	val rowL = row.toLong()

	for (k in 0 .. row) {
		built[k] = binomialDouble(rowL, k.toLong())
	}

	return built
}
